/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.component;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.awt.geom.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.io.*;

/**
 * <b>ZLine</b> is a simple ZShape implementation that uses a Line2D as the underlying
 * shape model.
 *
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @author  Jesse Grosjean
 */
public class ZLine extends ZShape {

    protected transient Line2D line;

    /**
     * ZLine constructor.
     */
    public ZLine() {
        super();
    }

    /**
     * ZLine constructor.
     */
    public ZLine(Line2D aLine) {
        this(aLine.getP1(), aLine.getP2());
    }

    /**
     * ZLine constructor.
     */
    public ZLine(Point2D pt1, Point2D pt2) {
        this(pt1.getX(), pt1.getY(), pt2.getX(), pt2.getY());
    }

    /**
     * ZLine constructor.
     */
    public ZLine(double x1, double y1, double x2, double y2) {
        super();
        setLine(x1, y1, x2, y2);
        fillPaint = null;
    }

    /**
     * Returns a clone of this object.
     *
     * @see ZSceneGraphObject#duplicateObject
     */
    protected Object duplicateObject() {
        ZLine newLine = (ZLine) super.duplicateObject();
        newLine.line = (Line2D) getLine().clone();
        return newLine;
    }
    /**
     * Return the line.
     */
    public Line2D getLine() {
        if (line == null) {
            line = new Line2D.Double();
        }
        return line;
    }

    /**
     * Return the current shape.
     */
    public Shape getShape() {
        return getLine();
    }

    /**
     * Set the line.
     */
    public void setLine(Line2D aLine) {
        setLine(aLine.getP1(), aLine.getP2());
    }

    /**
     * Set the line.
     */
    public void setLine(Point2D pt1, Point2D pt2) {
        setLine(pt1.getX(), pt1.getY(), pt2.getX(), pt2.getY());
    }

    /**
     * Set the line.
     */
    public void setLine(double x1, double y1, double x2, double y2) {
        getLine().setLine(x1, y1, x2, y2);
        reshape();
    }

    /////////////////////////////////////////////////////////////////////////
    //
    // Saving
    //
    /////////////////////////////////////////////////////////////////////////

    /**
     * Write out all of this object's state.
     * @param out The stream that this object writes into
     */
    public void writeObject(ZObjectOutputStream out) throws IOException {
        super.writeObject(out);

        Line2D line = getLine();
        Vector dimensions = new Vector();
        dimensions.add(new Double(line.getX1()));
        dimensions.add(new Double(line.getY1()));
        dimensions.add(new Double(line.getX2()));
        dimensions.add(new Double(line.getY2()));
        out.writeState("line", "line", dimensions);
    }

    /**
     * Set some state of this object as it gets read back in.
     * After the object is created with its default no-arg constructor,
     * this method will be called on the object once for each bit of state
     * that was written out through calls to ZObjectOutputStream.writeState()
     * within the writeObject method.
     * @param fieldType The fully qualified type of the field
     * @param fieldName The name of the field
     * @param fieldValue The value of the field
     */
    public void setState(String fieldType, String fieldName, Object fieldValue) {
        super.setState(fieldType, fieldName, fieldValue);

        if (fieldName.compareTo("line") == 0) {
            Vector dim = (Vector)fieldValue;
            double x1   = ((Double)dim.get(0)).doubleValue();
            double y1   = ((Double)dim.get(1)).doubleValue();
            double x2  = ((Double)dim.get(2)).doubleValue();
            double y2 = ((Double)dim.get(3)).doubleValue();

            setLine(x1, y1, x2, y2);
        }
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        out.defaultWriteObject();

        Line2D line = getLine();
        out.writeDouble(line.getX1());
        out.writeDouble(line.getY1());
        out.writeDouble(line.getX2());
        out.writeDouble(line.getY2());
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();

        double x1, y1, x2, y2;
        x1 = in.readDouble();
        y1 = in.readDouble();
        x2 = in.readDouble();
        y2 = in.readDouble();

        getLine().setLine(x1, y1, x2, y2);
    }
}
