/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.event;

import java.awt.*;
import java.awt.event.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.component.ZRectangle;
/**
 * <b>ZSelectionDeleteHandler</b> is a selection handler for use with a
 * <b>ZSelectionManager</b>.  <code>ZSelectionDeleteHandler</code> allows
 * the user to delete the current selection by pressing the DEL key.
 * <P>
 * <b>ZSelectionDeleteHandler</b> also allows registration of
 * <b>ZGroupListener</b> instances to be notified when items are deleted.
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @see edu.umd.cs.jazz.ZSelectionManager
 * @see ZCompositeSelectionHandler
 *
 * @author Antony Courtney, Yale University
 * @author Lance Good, Univesity of Maryland
 * @author Benjamin Bederson, University of Maryland
 */
public class ZSelectionDeleteHandler implements ZEventHandler, KeyListener, Serializable {
    /** true when event handler is active */
    private boolean active = false;

    /** true if this event handler active over all cameras */
    private boolean globallyActive = false;

    /** node this event handler attaches to */
    private ZNode node = null;

    /** the active camera for this handler */
    private ZCamera activeCamera = null;

    /** canvas this event handler attaches to */
    private ZCanvas canvas = null;

    /** registered event listeners */
    protected transient EventListenerList listenerList;

    /** key that deletes selected objects */
    private static final int deleteKey = KeyEvent.VK_DELETE;


    /**
     * Construct a new ZSelectionDeleteHandler.  Uses the canvas'
     * camera in obtaining selected nodes.
     * @param node The node this delete handler attaches to
     * @param canvas The canvas this handler is active on
     */
    public ZSelectionDeleteHandler(ZNode node, ZCanvas canvas) {
        this.node = node;
        this.canvas = canvas;
        listenerList = new EventListenerList();

        globallyActive = true;
    }

    /**
     * Construct a new ZSelectionDeleteHandler.
     * @param node The node this delete handler attaches to
     * @param camera The camera this delete handler attaches to
     * @param canvas The canvas this handler is active on
     */
    public ZSelectionDeleteHandler(ZNode node, ZCamera camera, ZCanvas canvas) {
        this(node,canvas);
        this.activeCamera = camera;
        globallyActive = false;
    }

    /**
     * Specifies whether this event handler is active
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
        if (this.active && !active) {
            // turn off this handler
            canvas.removeKeyListener(this);
            this.active = false;
        } else if (!this.active && active) {
            // turn on handler:
            this.active = true;
            canvas.addKeyListener(this);
        }
    }
    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
        return active;
    }

    /**
     * Add a ZGroupListener, registered for nodeRemoved events.
     */
    public void addGroupListener(ZGroupListener l) {
        listenerList.add(ZGroupListener.class, l);
    }

    /**
     * Remove a ZGroupListener, registered for nodeRemoved events.
     */
    public void removeGroupListener(ZGroupListener l) {
        listenerList.remove(ZGroupListener.class, l);
    }

    /**
     * Notifies all listeners that have registered interest for
     * notification on this event type.  The event instance
     * is lazily created using the parameters passed into
     * the fire method.  The listener list is processed in last to
     * first order.
     * @param id The event id (NODE_ADDED, NODE_REMOVED)
     * @param child The child being added or removed from this node
     * @see EventListenerList
     */
    protected void fireGroupEvent(int id, ZGroup parent, ZNode child) {
        if (listenerList == null) {
            return;
        }

        Object[] listeners = listenerList.getListenerList();

        ZGroupEvent e = null;
        if (id == ZGroupEvent.NODE_ADDED) {
            e = ZGroupEvent.createNodeAddedEvent(parent, child, false);
        } else if (id == ZGroupEvent.NODE_REMOVED) {
            e = ZGroupEvent.createNodeRemovedEvent(parent, child, false);
        }

        // Process the listeners last to first, notifying
        // those that are interested in this event
        for (int i = listeners.length-2; i>=0; i-=2) {
            if (listeners[i]==ZGroupListener.class) {
                switch (id) {
                case ZGroupEvent.NODE_ADDED:
                    ((ZGroupListener)listeners[i+1]).nodeAdded(e);
                    break;
                case ZGroupEvent.NODE_REMOVED:
                    ((ZGroupListener)listeners[i+1]).nodeRemoved(e);
                    break;
                }
            }
        }
    }

    /**
     * Key press event handler
     * @param <code>e</code> The event.
     */
    public void keyPressed(KeyEvent e) {
        int     keyCode = e.getKeyCode();

        if (keyCode==deleteKey) {
            ZNode node;
            ZNode handle;
            ZGroup parent;
            ArrayList selection;

            if (globallyActive) {
                selection = ZSelectionManager.getSelectedNodes(canvas.getRoot());
            }
            else {
                selection = ZSelectionManager.getSelectedNodes(activeCamera);
            }

            for (Iterator i=selection.iterator(); i.hasNext();) {
                node = (ZNode)i.next();
                handle = node.editor().getTop();
                parent = handle.getParent();
                parent.removeChild(handle);
                fireGroupEvent(ZGroupEvent.NODE_REMOVED, parent, node);
            }
        }
    }

    /**
     * Key release event handler
     * @param <code>e</code> The event.
     */
    public void keyReleased(KeyEvent e) {
    }

    /**
     * Key typed event handler
     * @param <code>e</code> The event.
     */
    public void keyTyped(KeyEvent e) {
    }
}