/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.event;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import javax.swing.event.*;
import java.util.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.component.ZRectangle;

/**
 * <b>ZSelectionScaleHandler</b> is a selection handler for use with a
 * <b>ZSelectionManager</b>.  <code>ZSelectionScaleHandler</code> allows
 * the user to scale the current selection using the PgUp / PgDn keys.
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @see edu.umd.cs.jazz.ZSelectionManager
 * @see ZCompositeSelectionHandler
 *
 * @author Antony Courtney, Yale University
 * @author Lance Good, University of Maryland
 * @author Benjamin Bederson, University of Maryland
 */
public class ZSelectionScaleHandler implements ZEventHandler, KeyListener, Serializable {
    /** true when event handler is active */
    private boolean active = false;

    /** true if event handler is globally active */
    private boolean globallyActive = false;

    /** The root of the scenegraph of the canvas this event handler watches */
    private ZNode root = null;

    /** Camera this event handler watches */
    private ZCamera activeCamera = null;

    /** canvas this event handler attaches to */
    private ZCanvas canvas = null;

    /** key to scale objects up a bit */
    private static final int scaleUpKey = KeyEvent.VK_PAGE_UP;
   
    /** key to scale objects down a bit */
    private static final int scaleDownKey = KeyEvent.VK_PAGE_DOWN;

    /** magnification factor for incremental scales */
    private static final double scaleDelta = 1.1f;

    /**
     * Construct a new ZSelectionScaleHandler.  This event handler operates
     * only on the canvas' camera.
     * @param node The node this event handler attaches to
     * @param canvas The canvas on which this event handler is active
     */
    public ZSelectionScaleHandler(ZCanvas canvas) {
	this.canvas = canvas;
	this.root = canvas.getRoot();
	globallyActive = true;
    }

    /**
     * Construct a new ZSelectionScaleHandler.  This event handler only
     * operates on the specified camera.
     * @param node The node this event handler attaches to
     * @param camera The camera on which this event handler operates
     * @param canvas The canvas on which this event handler is active
     */
    public ZSelectionScaleHandler(ZCamera camera, ZCanvas canvas) {
	this.canvas = canvas;
	this.activeCamera = camera;
    }

    /**
     * Specifies whether this event handler is active
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
	if (this.active && !active) {
	    // turn off this handler
	    canvas.removeKeyListener(this);
	    this.active = false;
	} else if (!this.active && active) {
	    // turn on handler:
	    this.active = true;
	    canvas.addKeyListener(this);
	}
    }
    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
	return active;
    }

    /**
     * Key press event handler
     * @param <code>e</code> The event.
     */
    public void keyPressed(KeyEvent e) {
	int     keyCode = e.getKeyCode();
	boolean scale = false;
	double scaleZ = 1.0;

	switch (keyCode) {
	case scaleUpKey:
	    scale = true;
	    scaleZ = scaleDelta;
	    break;
	case scaleDownKey:
	    scale = true;
	    scaleZ = 1.0 / scaleDelta;
	    break;
	}
	if (scale) {
	    Point2D pt1 = new Point2D.Double();

	    ArrayList selection;
	    if (globallyActive) {
		selection = ZSelectionManager.getSelectedNodes(root);
	    } else {
		selection = ZSelectionManager.getSelectedNodes(activeCamera);
	    }

	    ZTransformGroup transform = null;
	    ZBounds bounds = new ZBounds();
				// First, get bounds of all nodes
	    for (Iterator i=selection.iterator(); i.hasNext();) {
		ZNode node = (ZNode)i.next();
		bounds.add(node.getGlobalBounds());
	    }
				// Then, scale them around a common point
	    ZNode node;
	    Point2D center = bounds.getCenter2D();
	    for (Iterator i=selection.iterator(); i.hasNext();) {
		node = (ZNode)i.next();
				// Convert point and scale to node's coordinate system
		pt1.setLocation(center);
		node.globalToLocal(pt1);
				// Then, scale in node's local coordinate system
		transform = node.editor().getTransformGroup();
		transform.scale(scaleZ, pt1.getX(), pt1.getY());
	    }
	}
    }

    /**
     * Key release event handler
     * @param <code>e</code> The event.
     */
    public void keyReleased(KeyEvent e) {
    }

    /**
     * Key typed event handler
     * @param <code>e</code> The event.
     */
    public void keyTyped(KeyEvent e) {
    }
}
