/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
package edu.umd.cs.jazz.event;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.io.*;
import javax.swing.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;

/**
 * <b>ZoomEventhandler</b> provides event handlers for basic zooming
 * of a Jazz camera with the right button.  The interaction is that
 * the initial mouse press defines the zoom anchor point, and then
 * moving the mouse to the right zooms with a speed proportional
 * to the amount the mouse is moved to the right of the anchor point.
 * Similarly, if the mouse is moved to the left, the the camera is
 * zoomed out.
 *
 * <P>
 * This event handler will properly pan within internal cameras if
 * the clicked within an internal camera.
 *
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * @author  Benjamin B. Bederson
 */

public class ZoomEventHandler implements ZEventHandler, ZMouseListener, ZMouseMotionListener, Serializable {

    transient private ZCamera interactionCamera = null;
    transient private Point2D currentScreenPoint = null;
    transient private Point2D pressObjectPoint = null;
    transient private Point2D pressScreenPoint = null;
    private double minMag = 0.0;
    private double maxMag = -1.0;
    private boolean isZooming = false;
    private boolean active = false;             // True when event handlers are attached to a node
    private boolean globallyActive = false;
    private ZNode node = null;                  // The node the event handlers are attached to
    private ZCamera activeCamera = null;        // The camera we are attached to (if any)
    private ZoomEventState zoomEventState = null;

    private int allButton3Mask = (MouseEvent.BUTTON3_MASK |
                                  MouseEvent.ALT_GRAPH_MASK |
                                  MouseEvent.CTRL_MASK |
                                  MouseEvent.SHIFT_MASK);

    /**
     * Provides indirect access this event handlers private state.
     */
    protected class ZoomEventState {

        /**
         * Get the camera that is zooming.
         */
        public ZCamera getCamera() {return interactionCamera;}

        /**
         * Get event coords of current mouse posistion (in screen space).
         */
        public Point2D getCurrentScreenPoint() {return currentScreenPoint;}

        /**
         * Get event coords of mouse press (in object space).
         */
        public Point2D getPressScreenPoint() {return pressScreenPoint;}

        /**
         * Get event coords of mouse press (in object space).
         */
        public Point2D getPressObjectPoint() {return pressObjectPoint;}

        /**
         * Get the minimum allowed magnification.
         */
        public double getMaxMagnification() {return maxMag;}

        /**
         * Get the maximum allowed magnification (or disabled if less than 0)
         */
        public double getMinMagnification() {return minMag;}
    }

    /**
     * Constructs a new ZoomEventHandler.
     * @param node The node this event handler attaches to.
     */
    public ZoomEventHandler(ZNode node) {
        this(node, null);
        globallyActive = true;
    }

    /**
     * Constructs a new ZoomEventHandler.
     * @param node The node this event handler attaches to.
     * @param camera The camera to which this event handler attaches
     */
    public ZoomEventHandler(ZNode node, ZCamera camera) {
        this.node = node;
        this.activeCamera = camera;
        pressObjectPoint = new Point2D.Double();
        pressScreenPoint = new Point2D.Double();
        currentScreenPoint = new Point2D.Double();
        zoomEventState = new ZoomEventState();
    }

    /**
     * Specifies whether this event handler is active or not.
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
        if (this.active && !active) {
                                // Turn off event handlers
            this.active = false;
            node.removeMouseListener(this);
            node.removeMouseMotionListener(this);
        } else if (!this.active && active) {
                                // Turn on event handlers
            this.active = true;
            node.addMouseListener(this);
            node.addMouseMotionListener(this);
        }
    }

    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
        return active;
    }

    /**
     * Mouse press event handler
     * @param <code>e</code> The event.
     */
    public void mousePressed(ZMouseEvent e) {
        if ((e.getModifiers() & allButton3Mask) == MouseEvent.BUTTON3_MASK) {   // Right button only
            ZSceneGraphPath path = e.getPath();

            if (globallyActive) {
                interactionCamera = path.getCamera();
            }
            else {
                interactionCamera = activeCamera;
            }

            path.getTopCamera().getDrawingSurface().setInteracting(true);

            pressScreenPoint.setLocation(e.getX(), e.getY());
            currentScreenPoint.setLocation(pressScreenPoint);
            pressObjectPoint.setLocation(pressScreenPoint);

            path.screenToCamera(pressObjectPoint,interactionCamera);
            interactionCamera.cameraToLocal(pressObjectPoint,null);

            startZooming();
        }
    }

    /**
     * Mouse drag event handler
     * @param <code>e</code> The event.
     */
    public void mouseDragged(ZMouseEvent e) {
        if ((e.getModifiers() & allButton3Mask) == MouseEvent.BUTTON3_MASK) {   // Right button only
            currentScreenPoint.setLocation(e.getX(), e.getY());
        }
    }

    /**
     * Mouse release event handler
     * @param <code>e</code> The event.
     */
    public void mouseReleased(ZMouseEvent e) {
        if ((e.getModifiers() & allButton3Mask) == MouseEvent.BUTTON3_MASK) {   // Right button only
            stopZooming();

            e.getPath().getTopCamera().getDrawingSurface().setInteracting(false);

            if (interactionCamera != null) {
                                // do this to generate side effect camera event
                interactionCamera.setViewTransform(interactionCamera.getViewTransform());
            }

            interactionCamera = null;
        }
    }

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse has been clicked on a component.
     */
    public void mouseClicked(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse button has been moved on a node
     * (with no buttons no down).
     */
    public void mouseMoved(ZMouseEvent e) {
    }

    /**
     * Start animated zooming.
     */
    public void startZooming() {
        isZooming = true;
        zoomOneStep();
    }

    /**
     * Set the minimum magnification that the camera can be set to
     * with this event handler.  Setting the min mag to <= 0 disables
     * this feature.  If the min mag if set to a value which is greater
     * than the current camera magnification, then the camera is left
     * at its current magnification.
     * @param newMinMag the new minimum magnification
     */
    public void setMinMagnification(double newMinMag) {
        minMag = newMinMag;
    }

    /**
     * Set the maximum magnification that the camera can be set to
     * with this event handler.  Setting the max mag to <= 0 disables
     * this feature.  If the max mag if set to a value which is less
     * than the current camera magnification, then the camera is left
     * at its current magnification.
     * @param newMaxMag the new maximum magnification
     */
    public void setMaxMagnification(double newMaxMag) {
        maxMag = newMaxMag;
    }

    /**
     * Stop animated zooming.
     */
    public void stopZooming() {
        isZooming = false;
    }

    /**
     * Do one zooming step, sleep a short amount, and schedule the next zooming step.
     * This effectively continuously zooms while still accepting input events so
     * that the zoom center point can be changed, and zooming can be stopped.
     */
    public void zoomOneStep() {
        if (isZooming) {
                                // Now, generate and set next view transform.
            interactionCamera.setViewTransform(generateNextViewTransform(zoomEventState));

            try {
                                // The sleep here is necessary.  Otherwise, there won't be
                                // time for the primary event thread to get and respond to
                                // input events.
                Thread.sleep(20);

                                // If the sleep was interrupted, then cancel the zooming,
                                // so don't do the next zooming step
                SwingUtilities.invokeLater(new Runnable() {
                    public void run() {
                        ZoomEventHandler.this.zoomOneStep();
                    }
                });
            } catch (InterruptedException e) {
                isZooming = false;
            }
        }
    }

    /**
     * Return the next view transform for the interaction cameras zooming
     * sequence.
     *
     * @param state provides access to the current state of the event handler.
     * ZoomEventState is declared as a protected inner class in ZoomEventHandler.
     */
    protected AffineTransform generateNextViewTransform(ZoomEventState state) {
        AffineTransform result = interactionCamera.getViewTransform();

        double dx = (double)(state.getCurrentScreenPoint().getX() - state.getPressScreenPoint().getX());
        double scaleDelta = (1.0 + (0.001 * dx));

        double currentMag = state.getCamera().getMagnification();
        double newMag = currentMag * scaleDelta;

        if (newMag < minMag) {
            scaleDelta = minMag / currentMag;
        }
        if ((maxMag > 0) && (newMag > maxMag)) {
            scaleDelta = maxMag / currentMag;
        }

        Point2D pressPt = state.getPressObjectPoint();
        result.translate(pressPt.getX(), pressPt.getY());
        result.scale(scaleDelta, scaleDelta);
        result.translate(-pressPt.getX(), -pressPt.getY());

        return result;
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        pressObjectPoint = new Point2D.Double();
        pressScreenPoint = new Point2D.Double();
        currentScreenPoint = new Point2D.Double();
        isZooming = false;
    }
}