/**
 * @(#) ClosePackageCommand.java
 */

package venice;

import java.util.*;


/**
 * Closes the packages selected in Visualization. This means that all their children will 
 * become hidden and all relationships crossing the selected packages
 * will be combined in composite relationships not going in the package. This is same as
 * moving relationships up for each child.
 * 
 * @author	Petteri Kamppuri
 * @version	1.2
 */
public class ClosePackageCommand extends Command
{
    /**
     * This method iterates through the list of
     * selected <code>VModelElements</code> and does the close operation on each
     * package in the list.
     */
    public void execute()
    {
	Visualization myVisualization = getAppModule().getGui().getVisualization();
	
	/* Get current selection from visualisation subsystem*/
	VModelElement[] mySelection = myVisualization.getSelection();
	
	if(mySelection != null) {
	    // Iterate through selected nodes and handle all of type VPackage
	    for(int i = 0; i < mySelection.length; ++i) {
		if(mySelection[i] instanceof VPackage) {
		    VPackage myTempPackage = (VPackage) mySelection[i];
		    
		    moveAllChildRelationshipsToThisPackage(myTempPackage);
		    myTempPackage.setShowContent(false);
		    myTempPackage.setState(VPackage.CLOSED);
		}
	    }
	}	
    }
    
    /**
     * This method closes the given <code>VPackage</code>.
     * Closing means hiding all children and moving all their relationships
     * up to the level of the package.
     * 
     * <pre>
     * closePackage(inPackage)
     * 		for each child as myChild
     *              for each relationship of myChild and its children as myRel
     *                      // handle relationship leaving this package
     *                      if not (myRel.src isDescentOf this and myRel.dst isDescentOf this)
     *                              myRel.hide/remove from visualization
     *                              inPackage.addToCompositeRelationship(myRel)
     *              myChild.hide // or global hide all children with ZInvisibleGroup
     *      inPackage.setShowContents(false)
     * </pre>
     * 
     * @param inPackage The package to close.
     */
    public static void moveAllChildRelationshipsToThisPackage(VPackage inPackage)
    {
	Visualization myVisualization = getAppModule().getGui().getVisualization();
	VModelElement[] myChildren = inPackage.getVChildren();
	
	// Iterate through children and for each child get relationships and put
	// them to a composite relationship and remove them from the visualization
	for(int j = 0; j < myChildren.length; ++j) {
	    LinkedList myList = myChildren[j].getRelationships(inPackage);
	    if(myList != null) {
		ListIterator myIterator = myList.listIterator(0);
		
		while(myIterator.hasNext()) {
		    VRelationship myRel = (VRelationship) myIterator.next();
		    
		    addRelationshipToCompRelListNotifyingOtherEnd(inPackage, myRel);

		    myVisualization.removeVRelationship(myRel);
		}
	    }
	}
    }

    /**
     * @param inElement the element to which the <code>inRel</code> will be added as a composite relationship
     * @param inRel the relationship which will be placed as a composite relationship
     */
    public static void addRelationshipToCompRelListNotifyingOtherEnd(VModelElement inElement, VRelationship inRel)
    {
	VRelationship myNewCompRel = inElement.addToCompositeRelationshipList(inRel);
	
	if(myNewCompRel != null) {
	    // Add relationship to the _other_ end
	    if(myNewCompRel.getSource() == inElement) {
		//inRel.getDestination().addToCompositeRelationshipList(myNewCompRel);
		inRel.getDestination().getCompositeRelationships().add(myNewCompRel);
		//System.out.println("Added relationship to the other ends comp rels " + inRel.getDestination().getId());
	    } else {
		//inRel.getSource().addToCompositeRelationshipList(myNewCompRel);
		inRel.getSource().getCompositeRelationships().add(myNewCompRel);
		//System.out.println("Added relationship to the other ends comp rels 2 " + inRel.getSource().getId());
	    }
	}
    }
}
