 
package venice;

/**
 * <code>Command</code> is an abstract base class for a hierarchy of commands.
 * They implement the idea behind the design pattern Command.
 * This base class stores a reference to the current instance
 * of <code>AppModule</code> as a class variable, so it can be used in various
 * subclasses without being given as a parameter.
 * 
 * <p><code>Command</code> is a part of the design pattern Command.
 * GUI and other objects have instances of <code>Command</code> subclasses.
 * When necessary, these objects can <code>execute</code> the <code>Command</code>
 * instance they have to perform activities. This is a nice of way of abstracting
 * the actual actions and actors of the objects that initiate these actions.</p>
 * 
 * <p>Subclasses probably should have a constructor that takes as
 * parameters all references (to other objects) that are needed for that
 * particular command to do its duties.</p>
 * 
 * <p><code>Command</code> has a static <code>setAppModule</code> method,
 * which can be used to set the static AppModule reference to point to the
 * current AppModule instance. Subclasses can use <code>getAppModule</code> to get
 * this reference to access any data they need in the system.</p>
 * 
 * @author	Petteri Kamppuri
 * @version	1.0
 * 
 */
public abstract class Command
{
    /**
     * This abstract method is overridden in subclasses. The overridden
     * version does whatever the <code>Command</code> is supposed to do.
     * 
     * <p>Users of these subclasses
     * can rely on the <code>Command</code> interface, and just
     * call <code>execute()</code> when they want the given
     * subclass to perform an operation.
     */
    public abstract void execute();
    
    /**
     * Takes the given object and puts a reference to that
     * in the <code>stAppModule</code> class variable.
     * Subclasses of <code>Command</code> can then use
     * the saved <code>AppModule</code> reference for their
     * purposes.
     * 
     * @param inAppModule	the reference to save
     */
    public static void setAppModule(AppModule inAppModule)
    {
	if(inAppModule != null)
	    stAppModule = inAppModule;
    }
    
    /**
     * Returns a reference to the saved <code>stAppModule</code> class variable.
     * Subclasses of <code>Command</code> can then use
     * the saved <code>AppModule</code> reference for their
     * purposes.
     * 
     * @param inAppModule	the reference to save
     */
    protected static AppModule getAppModule()
    {
	if(stAppModule == null)
	    throw new RuntimeException("Use of AppModule.getAppModule before initialized!");
	
	return stAppModule;
    }
    
    /**
     * A class variable to hold a reference to the current
     * instance of <code>AppModule</code>. Subclasses can
     * take advantage of this reference.
     */
    private static AppModule	stAppModule = null;
}










