package venice;

import javax.swing.*;
import java.awt.*;
import java.io.*;

/**
 * Utility class containing error handling dialog functions and 
 * functions for showing about and license dialogs.
 *
 * @author Hannu Laurila
 */
public class Util
{
    private static final String stAboutHTML =
        "<html><body>" +
        "<center>" +
        "<h1>VENICE</h1><p>" +
        "UML Visualization tool for visualizing<br>" +
        "software architectures<br>" +
        "<p>" +
        "Copyright 2001 Department of Computer Science,<br>" +
        "University of Helsinki<br>" +
        "<p>" +
	"Venice Team:<br>" +
	"Petteri Kamppuri<br>" +
	"Hannu Laurila<br>" +
	"Sami Ilonen<br>" +
	"Antti Pietarinen<br>" +
	"Claudio Riva (client, instructor)<br>" +	
	"<p>" +
        "JAZZ Zoomable User Interface library<br>" +
        "Copyright 1998-2001 University of Maryland<br>" +
        "<p>" +
	//        "Insert JAXP, Crimson and Xalan copyrights here." +
        "</center>" + 
        "</body></html>";

    // FIXME: Copyrights above.

    /**
     * private constructor prevents instantiation 
     */
    private Util() 
    {
        // Prevent instantiation
    }

    /**
     * shows an error message dialog with first a short error description
     * in a text label and then a long error description in a text area.
     *
     * @param myParent parent component for which to make the dialog modal.
     *                 May be null.
     * @param myShortErr short error description
     * @param myLongErr longer error description (a dump). May be null.
     */
    public static void handleError (Component myParent, 
                                        String myShortErr, 
                                        String myLongErr) {
        Object[] myErrorMessages;
        
        if (myLongErr != null) {
            JTextArea myLongerrTextArea = new JTextArea(myLongErr);
            myErrorMessages = new Object[] { myShortErr,
                                             myLongErr };            
        } else {
            myErrorMessages = new Object[] { myShortErr };
        }
        
        JOptionPane.showMessageDialog(myParent, 
                                      myErrorMessages, 
                                      "Error!", 
                                      JOptionPane.ERROR_MESSAGE);
    }

    /**
     * Creates an error message dialog to inform user about exception.
     *
     * @param myParent parent component for which to make the dialog modal.
     * @param myEx exception to inform user about.
     */
    public static void handleException(Component myParent, Exception myEx)
    {
        System.out.println("--- BEGIN EXCEPTION STACK TRACE ---");
        myEx.printStackTrace();
        System.out.println("--- END EXCEPTION STACK TRACE ---");

        String myMsg = 
            "The following error was encountered: " + myEx.toString();
        handleError(myParent, myMsg, null);
    }

    /**
     * Creates an error message dialog to inform user about exception.
     */
    public static void handleException(Exception inEx)
    {
        handleException(null, inEx);
    }

    public static void showAboutDialog(JFrame inParent)
    {
        JLabel myAboutLabel = new JLabel(stAboutHTML);
        JOptionPane.showMessageDialog(inParent,
                                      myAboutLabel,
                                      "About VENICE",
                                      JOptionPane.PLAIN_MESSAGE);
    }

    public static void showLicenseDialog(JFrame inParent)
    {
        // string buffer for storing the license text read from file.
	StringBuffer myLicense = new StringBuffer();
	    
        // Read license text from jar archive.
	try {
	    InputStream licenseInputStream = 
		(new Util()).getClass().getClassLoader().
		getResourceAsStream("venice/gpl.txt");
	    
	    BufferedReader myRdr = 
		new BufferedReader(new InputStreamReader(licenseInputStream));
	    
	    String myLine;
	    
	    while ((myLine=myRdr.readLine())!=null) {
		myLicense.append(myLine + "\n");
	    }
	    
	    myRdr.close();
	}
	catch (Exception myEx) {	    
            // file could not be read. Use hard-coded text.
	    System.out.println("gpl.txt could not be read from archive");
	    myLicense = 
                new StringBuffer("This software is distributed under " +
                                 "GNU GENERAL PUBLIC LICENSE, VERSION 2.");
	}
	

	JTextArea myLicenseArea = new JTextArea(myLicense.toString());
	myLicenseArea.setFont(new Font("Monospaced", 
				       Font.PLAIN,
				       12));
	myLicenseArea.setEditable(false);
        myLicenseArea.setCaretPosition(0);

	JScrollPane myScrollPane = new JScrollPane(myLicenseArea);

	Dimension myDimension = new Dimension(600,400);
	myScrollPane.setMaximumSize(myDimension);
	myScrollPane.setPreferredSize(myDimension);


        JOptionPane.showMessageDialog(inParent,
                                      myScrollPane,
                                      "VENICE License information",
                                      JOptionPane.PLAIN_MESSAGE);
    }

    /**
     * Causes command to be executed later using SWING Event thread.
     * Swing and Jazz are not thread safe so everything should be
     * done in event dispatch thread to prevent concurrent access.
     */
    public static void invokeCommandLater(Command inCmd)
    {
	final Command myCmd = inCmd;
	Runnable myCmdRunner = new Runnable() {
		public void run()
		{
		    myCmd.execute();
		}
	    };

	SwingUtilities.invokeLater(myCmdRunner);
    }

    /**
     * test method for testing about and license dialog.
     */
    public static void main(String[] args)
    {
        showAboutDialog(null);
        showLicenseDialog(null);
    }
}
