/**
 * Copyright (C) 1998-2000 by University of Maryland, College Park, MD 20742, USA
 * All rights reserved.
 */
//package edu.umd.cs.jazz.event;

package venice;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.io.*;
import javax.swing.event.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.ZRectangle;

/**
 * <b>VSelectionModifyHandler</b> is a <b>ZEventHandler</b> that allows items
 * to be selected with the mouse, and maintains the current selection.
 *
 * As an event handler, <b>ZSelectionManager</b> directly supports the
 * following actions:
 * <ul>
 * <li> Click to select/unselect an item.
 * <li> Click-and-drag on the background to marquee select.
 * </ul>
 * Holding down the <b>SHIFT</b> key while performing the above actions
 * will extend the current selection with the newly selected items, otherwise
 * the previous selection (if any) will be replaced by the new selection.
 * <P>
 * <b>Warning:</b> Serialized and ZSerialized objects of this class will not be
 * compatible with future Jazz releases. The current serialization support is
 * appropriate for short term storage or RMI between applications running the
 * same version of Jazz. A future release of Jazz will provide support for long
 * term persistence.
 *
 * <p>
 * <b>THIS IS A MODIFIED VERSION OF ZSELECTIONMODIFYHANDLER FROM JAZZ CODE!</b>
 *
 * @author Antony Courtney, Yale University
 * @author Lance Good, University of Maryland
 * @author Benjamin Bederson, University of Maryland
 * @author modified by Antti Pietarinen
 */
public class VSelectionModifyHandler implements ZEventHandler, ZMouseListener, ZMouseMotionListener, Serializable {
    /** true when event handler is active */
    private boolean active = false;

    /** true if event handler is globally active */
    private boolean globallyActive = false;

    /** node this event handler attaches to */
    private ZNode node = null;

    /** camera this event handler watches */
    private ZCamera activeCamera = null;

    /** the camera with which this handler is interacting */
    private ZCamera interactionCamera = null;

    /** canvas this event handler attaches to */
    private ZCanvas canvas = null;

    /** parent node to use for marquee selection */
    private ZGroup marqueeLayer;

    private ArrayList   prevMotionSelection; // Reusable list for implementing marquee selection
    private ArrayList   itemsToRemove;       // Reusable list for implementing marquee selection
    private ZNode       selNode;             // Selected object
    private boolean     moved;               // Was there a move between mouse press and release
    private boolean     consumed;            // Was the mouse press consumed
    private transient Point2D     pt1, pt2;            // Utiltity points used temporarily within event handlers
    private transient Point2D     pressPt;             // Event coords of mouse press event (in window coordinates)
    private transient Point2D     dragPt;              // Event coords of mouse drag event (in window coordinates)
    private transient Point2D     prevPt;              // Event coords of previous mouse event (in window coordinates)
    private ZVisualLeaf marquee;             // Rectangle node representing marquee selection

    private int allButton1Mask   = (MouseEvent.BUTTON1_MASK |
                                    MouseEvent.ALT_GRAPH_MASK |
                                    MouseEvent.CTRL_MASK |
                                    MouseEvent.SHIFT_MASK);

    private int allButton1ButShiftMask   = (MouseEvent.BUTTON1_MASK |
                                            MouseEvent.ALT_GRAPH_MASK |
                                            MouseEvent.CTRL_MASK);

    private int allButton3ButShiftMask   = (MouseEvent.BUTTON3_MASK |
                                            MouseEvent.ALT_GRAPH_MASK |
                                            MouseEvent.CTRL_MASK);

    /**
     * Construct a new VSelectionModifyHandler. This event handler will be
     * active on all cameras.
     * @param node The node this event handler attaches to.
     * @param canvas The canvas this event handler attaches to
     * @param marqueeLayer The layer to draw the marquee on
     */
    public VSelectionModifyHandler(ZNode node, ZCanvas canvas, ZGroup marqueeLayer) {
        this.node = node;
        this.canvas = canvas;
        this.marqueeLayer = marqueeLayer;

        pt1 = new Point2D.Double();
        pt2 = new Point2D.Double();
        pressPt = new Point2D.Double();
        dragPt = new Point2D.Double();
        prevPt = new Point2D.Double();
        prevMotionSelection = new ArrayList();
        itemsToRemove = new ArrayList();
        marquee = null;

        globallyActive = true;
    }

    /**
     * Construct a new VSelectionModifyHandler.  This event handler will be
     * active on the specified camera.
     * @param node The node this event handler attaches to.
     * @param camera The camera to which this event handler attaches.
     * @param canvas The canvas this event handler attaches to
     * @param marqueeLayer The layer to draw the marquee on
     */
    public VSelectionModifyHandler(ZNode node, ZCamera camera, ZCanvas canvas, ZGroup marqueeLayer) {
        this(node,canvas,marqueeLayer);
        this.activeCamera = camera;
        globallyActive = false;
    }

    /**
     * Specifies whether this event handler is active
     * @param active True to make this event handler active
     */
    public void setActive(boolean active) {
        if (this.active && !active) {
            // turn off this handler, and all registered handlers.
            this.active = false;
            node.removeMouseListener(this);
            node.removeMouseMotionListener(this);

            if (globallyActive) {
                ZSelectionManager.unselectAll(canvas.getRoot());
            }
            else {
                ZSelectionManager.unselectAll(activeCamera);
            }
        } else if (!this.active && active) {
            this.active = true;
            node.addMouseListener(this);
            node.addMouseMotionListener(this);
            canvas.requestFocus();
        }
    }

    /**
     * Determines if this event handler is active.
     * @return True if active
     */
    public boolean isActive() {
        return active;
    }

    /**
     * Returns the node to which this selection manager is attached.
     * Primarily for use by associated selection handlers.
     */
    public ZNode getNode() {
        return node;
    }

    /**
     * Returns the camera on which this event handler is watching for
     * selection change
     */
    public ZCamera getCamera() {
        return activeCamera;
    }

    /**
     * Returns the canvas to which this selection manager is attached.
     * Primarily for use by associated selection handlers.
     */
    public ZCanvas getCanvas() {
        return canvas;
    }

    /**
     * Specify the node that the selection "marquee" should be put on.
     * The marquee is the rectangle that the user drags around to select things within.
     * @param layer The node that the marquee should be put under
     */
    public void setMarqueeLayer(ZGroup layer) {
        marqueeLayer = layer;
    }

    /**
     * Determine the "marquee" layer that is used for selection.
     * @return layer The node that the marquee is put under
     */
    public ZGroup getMarqueeLayer() {
        return marqueeLayer;
    }

    /**
     * Mouse press event handler
     * @param <code>e</code> The event.
     */
    public void mousePressed(ZMouseEvent e) {
        if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK || 
	    (e.getModifiers() & allButton3ButShiftMask) == MouseEvent.BUTTON3_MASK ) {
            // Left button only

            ZSceneGraphPath path = e.getPath();
            ZCamera topCamera = path.getTopCamera();
            boolean marqueeSelect = false;
            moved = false;

            topCamera.getDrawingSurface().setInteracting(true);

            if (globallyActive) {
                interactionCamera = path.getCamera();
            }
            else {
                interactionCamera = activeCamera;
            }

            // If no object has been picked, than the user has clicked on
            // the camera - so we set the selNode to null

            if (path.getObject() != null) {
                selNode = path.getNode();
            }
            else {
                selNode = null;
            }

            pressPt.setLocation(e.getX(), e.getY());
            prevPt.setLocation(pressPt);

            if (selNode == null) {
                                // If object is not pressed on
                if (!e.isShiftDown()) {
                    if (globallyActive) {
                        ZSelectionManager.unselectAll(canvas.getRoot());
                    }
                    else {
                        ZSelectionManager.unselectAll(interactionCamera);
                    }
                }

                consumed = true;
                e.consume();

                marqueeSelect = true;
            } else {
                                // Else, object pressed on
                if (!e.isShiftDown()) {
                                // Shift key not down
                    if (!ZSelectionManager.isSelected(selNode)) {
                        if (globallyActive) {
                            ZSelectionManager.unselectAll(canvas.getRoot());
                        }
                        else {
                            ZSelectionManager.unselectAll(interactionCamera);
                        }
                        ZSelectionManager.select(selNode);
                    }
                }
            }

            if (marqueeSelect) {
                double dz;
                ZRectangle rect;
                pt1.setLocation(pressPt);

                // Find the correct pen width and position for the marquee
                // rectangle

                if (globallyActive) {
                    dz = topCamera.cameraToLocal(pt1, marqueeLayer);
                }
                else {
                    dz = path.screenToCamera(pt1,interactionCamera);
                    dz *= interactionCamera.cameraToLocal(pt1, marqueeLayer);
                }

                rect = new ZRectangle(pt1.getX(), pt1.getY());
                rect.setPenWidth(1.0 * dz);
                rect.setPenPaint(Color.black);
                rect.setFillPaint(null);
                marquee = new ZVisualLeaf(rect);
                marquee.setPickable(false);
                marquee.setFindable(false);
                marqueeLayer.addChild(marquee);
            }
        }
    }

    /**
     * Mouse drag event handler
     * @param <code>e</code> The event.
     */
    public void mouseDragged(ZMouseEvent e) {
        if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK) {
            // Left button only

            ZSceneGraphPath path = e.getPath();
            ZCamera topCamera = path.getTopCamera();
            moved = true;

            dragPt.setLocation(e.getX(), e.getY());
            if ((marquee != null) && (interactionCamera != null)) {
                                // Select by marquee:
                                // First, modify marquee

                ZBounds marqueeRect = new ZBounds();

                pt1.setLocation(dragPt);
                pt2.setLocation(pressPt);

                if (globallyActive) {

                    // First put the points in the marquees coordinates

                    topCamera.cameraToLocal(pt1, marquee);
                    topCamera.cameraToLocal(pt2, marquee);

                    // Create the initial marquee rectangle

                    marqueeRect.setFrameFromDiagonal(pt1,pt2);

                    // Now we need to intersect with the cameras bounds to
                    // make sure we don't try to select outside the camera

                    ZBounds cameraBounds = interactionCamera.getBounds();
                    AffineTransform cameraTransform = path.getCameraTransform();
                    cameraBounds.transform(cameraTransform);
                    topCamera.cameraToLocal(cameraBounds,marquee);

                    Rectangle2D.intersect(marqueeRect,cameraBounds,marqueeRect);
                    ZRectangle rect = (ZRectangle)marquee.getFirstVisualComponent();
                    // Set the visible marquee with our intersected rect

                    rect.setRect((ZBounds)marqueeRect.clone());

                    // Now put the marquee rectangle into the cameras global coordinates
                    topCamera.localToCamera(marqueeRect,marquee);
                    path.screenToCamera(marqueeRect,interactionCamera);
                    interactionCamera.cameraToLocal(marqueeRect,null);
                }
                else {

                    // First put the points in the cameras coordinates

                    path.screenToCamera(pt1,interactionCamera);
                    path.screenToCamera(pt2,interactionCamera);

                    // Now from the camera to the marquee

                    interactionCamera.cameraToLocal(pt1,marquee);
                    interactionCamera.cameraToLocal(pt2,marquee);

                    // Create the initial marquee rectangle and intersect it
                    // with the camera view bounds

                    marqueeRect.setFrameFromDiagonal(pt1,pt2);
                    ZRectangle rect = (ZRectangle)marquee.getFirstVisualComponent();
                    ZBounds viewBounds = interactionCamera.getViewBounds();

                    marquee.globalToLocal(viewBounds);
                    Rectangle2D.intersect(marqueeRect,viewBounds,marqueeRect);

                    // Set the visible marquee with our intersected rect

                    rect.setRect((ZBounds)marqueeRect.clone());

                    // Now put the marquee rectangle into the cameras global coordinates
                    marquee.localToGlobal(marqueeRect);
                }


                ZFindFilter filter = new ZMagBoundsFindFilter(marqueeRect, interactionCamera.getMagnification());
                ZNode node;
                ArrayList nodes = interactionCamera.findNodes(filter);
                for (Iterator i=nodes.iterator(); i.hasNext();) {
                    node = (ZNode)i.next();

                    if (!prevMotionSelection.contains(node)) {
                        if (e.isShiftDown()) {
                            if (ZSelectionManager.isSelected(node)) {
                                ZSelectionManager.unselect(node);
                            }
                            else {
                                ZSelectionManager.select(node);
                            }
                            prevMotionSelection.add(node);
                        }
                        else {
                            ZSelectionManager.select(node);
                            prevMotionSelection.add(node);
                        }
                    }
                }
                                // Unselect items that don't overlap region, but previously did
                                // Make a list of these items to safely remove from list afterwards
                itemsToRemove.clear();
                for (Iterator i=prevMotionSelection.iterator(); i.hasNext();) {
                    node = (ZNode)i.next();
                    if (!nodes.contains(node)) {
                        if (e.isShiftDown()) {
                            if (ZSelectionManager.isSelected(node)) {
                                ZSelectionManager.unselect(node);
                            }
                            else {
                                ZSelectionManager.select(node);
                            }
                            itemsToRemove.add(node);
                        }
                        else {
                            ZSelectionManager.unselect(node);
                            itemsToRemove.add(node);
                        }
                    }
                }

                for (Iterator i=itemsToRemove.iterator(); i.hasNext();) {
                    node = (ZNode)i.next();
                    prevMotionSelection.remove(node);
                }

                e.consume();
            }
        }
    }

    /**
     * Mouse release event handler
     * @param <code>e</code> The event.
     */
    public void mouseReleased(ZMouseEvent e) {
        if ((e.getModifiers() & allButton1ButShiftMask) == MouseEvent.BUTTON1_MASK || 
	    (e.getModifiers() & allButton3ButShiftMask) == MouseEvent.BUTTON3_MASK ) {
            // Left button only

            ZSceneGraphPath path = e.getPath();

            if (!moved) {
                if (selNode != null && e.isShiftDown()) {
                    // Shift key down
                    if (ZSelectionManager.isSelected(selNode)) {
                        ZSelectionManager.unselect(selNode);
                    } else {
                        ZSelectionManager.select(selNode);
                    }
                }
            }

            selNode = null;

            prevMotionSelection.clear();

            if (marquee != null) {
                marqueeLayer.removeChild(marquee);
                marquee = null;
            }

            if (consumed) {
                consumed = false;
                e.consume();
            }

            path.getTopCamera().getDrawingSurface().setInteracting(false);
        }
    }

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse has been clicked on a component.
     */
    public void mouseClicked(ZMouseEvent e) {
    }

    /**
     * Invoked when the mouse button has been moved on a node
     * (with no buttons no down).
     */
    public void mouseMoved(ZMouseEvent e) {
    }
}
