package venice;

import java.io.*;
import javax.swing.*;

/**
 * Visualization represents a single visualization of some data.
 * It's used by getting an instance of a class which implemenets
 * the Visualization interface form the VisualizationFactory.
 * After an instance has been acquired, one can add different
 * VModelElements to the visualization (see subclasses of
 * VModelElement).
 * <p>
 * Note that different VModelElement subclasses (e.g. VPackage)
 * returned by the factory methods of this interface are compatible
 * only with the particular visualization from whose instance
 * they have been acquired.
 * <p>
 * Visualization supports a selection listener (a VSelectionListener) which
 * is notified when the selection in the visualization area changes.
 *
 * @author Hannu Laurila
 */

public interface Visualization
{
    /**
     * a constant describing that there is no active mode 
     */
    public static final int MODE_NONE=0;

    /**
     * a constant describing panning mode (moving the visualization area)
     */
    public static final int MODE_PAN=1;

    /**
     * a constant describing a mode where things can be selected.
     */
    public static final int MODE_SELECT=3;

    /**
     * a constant describing a mode where the visualization can be zoomed.
     */
    public static final int MODE_ZOOM=2;

    /**
     * Adds a relationship to the visualization. 
     * Relationships is either a realization, a generalization
     * or a dependency. 
     * The given <code>VRelationship</code> must have
     * its source and destination set. If they are not set,
     * the relationship will not be added. Also adds a reference
     * of the given relationship to both its source and destination
     * node.
     * 
     * @param inRelship relationship to add to visualization.
     */
    public void addVRelationship(VRelationship inRelship);

    /**
     * Add the given relationship to the visualization.
     * The relationship must have its source and destination node
     * set.
     * 
     * @param inRelship relationship to add to visualization.
     */
    public void addVRelationshipWithoutAddingOriginalReference(VRelationship inRelship);

    /**
     * findElementById finds a model element stored in visualization
     * subsystem data structures by its id.
     *
     * @param inId ID of the model element to find
     * @return reference to the element if found, null otherwise
     */
    public VModelElement findElementById(String inId);

    /**
     * getDebugMenu returns a menu which can be placed to a menubar.
     * Debug menu includes menu items for getting debug information
     * regarding the visualization. The information is mainly ment for
     * developers
     */
    public JMenu getDebugMenu();

    /**
     * returns a reference to the root package of the visualization.
     */
    public VPackage getRootPackage();

    /**
     * returns currently selected items in the visualization
     *
     * @return array consisting of currently selected elements
     */
    public VModelElement[] getSelection();

    /**
     * returns a Swing component representing the visualization.
     * The swing component can then be placed on the user interface
     * of the visualization software
     *
     * @return a reference to a Swing JComponent.
     */
    public JComponent getSwingComponent();

    /**
     * gets the current mode of the visualization area.
     * Constants beginning with a MODE_ describe the possible
     * visualization area modes.
     */
    public int getToolMode();

    /**
     * Load layout information from a stream
     *
     * @param layoutStream Layout stream which has been saved using
     *                     saveLayout-method.
     */    
    public void loadLayout(InputStream layoutStream)
        throws IOException;

    /**
     * constructs a default layout for the visualization.
     * What is 'the default layout' depends on the particular
     * visualization subsystem implementation used.
     *
     * This method should be called after all model elements loaded
     * from the model data source have been inserted to the 
     * visualization.
     */
    public void makeDefaultLayout();

    /**
     * removes element specified by inId from lookup structures.
     * After removeElement the element can not be looked up using 
     * findElementById. Note that this does not remove the element
     * from any VPackage if it is there.
     */
    public void removeElement(String inId);

    /**
     * a factory method for generating new VComponents.
     */
    public VComponent newVComponent(String inId);

    /**
     * a factory method for generating new VDependencies.
     */
    public VDependency newVDependency(String inId);

    /**
     * a factory method for generating new VGeneralizations.
     */
    public VGeneralization newVGeneralization(String inId);

    /**
     * a factory method for generating new VInterfaces.
     */
    public VInterface newVInterface(String inId);

    /**
     * a factory method for generating new VPackages.
     */
    public VPackage newVPackage(String inId);

    /**
     * a factory method for generating new VRealizations.
     */
    public VRealization newVRealization(String inId);

    /**
     * remove a relationship from the visualization.
     */
    public void removeVRelationship(VRelationship inRelship);

    /**
     * Saves the layout information to an output stream.
     *
     * Layout is saved in a format which can be loaded using 
     * loadLayout() method.
     *
     * @param layoutStream a stream where to save to layout information
     */    
    public void saveLayout(OutputStream layoutStream)
        throws IOException;

    /**
     * sets the currently selected model elements in the visualization.
     */
    public void setSelection(VModelElement[] inSelection);

    /**
     * sets the object which receives notifications about selection changes.
     *
     * The object must implement the VSelectionListener interface. After
     * been notified, the object can query the currently selected 
     * model elements via getSelection method.
     *
     * @param inListener a reference to a VSelectionListener which will
     *        receive notifications about changes in selection.
     */
    public void setSelectionListener(VSelectionListener inListener);

    /**
     * sets the current operating mode of the visualization area.
     *
     * @param inMode one of constants beginning with a MODE_ keyword.
     */
    public void setToolMode(int inMode);

    /**
     * sets the current speed of zooming
     *
     * @param inDx FIXME: What is this really
     */
    public void setZoomVolume(int inDx);

    /**
     * Zooms the visualization area to a particular element.
     *
     * zooms the visualization area so that inElement is as big as possible
     * but still fully visible.
     *
     * @param inElement the model element to be zoomed on.
     */
    public void zoomTo(VModelElement inElement);

    /**
     * Zooms to the current selection on the visualization area.
     *
     * The currently selected model elements will be visible.
     */
    public void zoomToSelection();

    /**
     * Clears the visualization (drops everything from the root package).
     * <p>
     * Note that VPackages which were in the visualization lose their childs 
     * so don't rely on getVChildren() after clearing the visualization.
     * In general, the VModelElement instances in the visualization
     *  should not be used after the visualization has been cleared.
     */
    public void clear();

    /**
     * enters non-interactive mode. 
     * <p>
     * In non-interactive mode processing
     * speed is gained by not updating the visualization area constantly.
     * <p>
     * Non-interactive mode should be entered before doing heavy tasks
     * like setting up the initial view.
     * <p>
     * No user interaction should be allowed during non-interactive mode.
     */
    public void enterNonInteractiveMode();

    /**
     * leaves non-interactive mode.
     */
    public void leaveNonInteractiveMode();

    /**
     * brings up a print dialog which user can use to send currently
     * visible view to printer.
     */
    public void printCurrentView();
}
