/**
 * @(#) ZInterface.java
 */

package venice;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.geom.Ellipse2D;
import java.util.LinkedList;
import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.component.*;
import edu.umd.cs.jazz.util.*;

/**
 * ZInterface is a concrete implementation of VInterface using
 * Jazz library.
 *
 * @author Hannu Laurila
 * @author Antti Pietarinen
 */
public class ZInterface extends ZVisualLeafModelElement implements VInterface
{
    /** size of the circle in pixels from left to right (not radius!) */
    private static final double CIRCLE_SIZE = 15;

    /** type of the interface, one of the TYPE_ constants */
    private int mInterfaceType;    

    /** view mode of the interface, one of the VIEWMODE_constants. */
    private int mViewMode;

    /**
     * operations of the interface as an array. toString() method is
     * used to get a String representation
     */
    private Object[] mOperations;    

    // Circle representation:

    /** visual component representing the label under the circle */
    private ZLabel mInterfaceLabel;
    
    /** visual component representing the circle */
    private ZEllipse mInterfaceEllipse;

    /** visual component representing the operations list text area */
    private ZText mExpandedInterfaceOpsText;

    /** visual component representing the <<interface>> label */
    private ZLabel mExpandedInterfaceStereoLabel;

    /** visual component representing the label for interface */
    private ZLabel mExpandedInterfaceLabel;

    /** visual component representing the class rectangle */
    private ZRectangle mExpandedInterfaceRectangle;

    /** visual component representing the line between name and operations */
    private ZLine mExpandedInterfaceSeparator;

    /** list of dependencies greated to substitute this interface when hidden.*/
    private LinkedList mSubsDeps;

    /** list of relationships to this interface which are hidden currently */
    private LinkedList mHiddenRelationships;

    /**
     * Constructor.
     *
     * @param inId model element id
     */
    public ZInterface(String inId)
    {
	super(inId);

	this.mHiddenRelationships = new LinkedList();
	this.mSubsDeps = new LinkedList();
	this.mInterfaceType = TYPE_OFFERED;
	this.mViewMode = VIEWMODE_CIRCLE;
	this.mOperations = new String[0];

	this.mInterfaceLabel = new ZLabel();
	this.mInterfaceEllipse = new ZEllipse(0,0,CIRCLE_SIZE,CIRCLE_SIZE);
        this.mExpandedInterfaceOpsText = new ZText();
        this.mExpandedInterfaceOpsText.setEditable(false);
        this.mExpandedInterfaceLabel = new ZLabel();
        this.mExpandedInterfaceStereoLabel = new ZLabel();
        this.mExpandedInterfaceRectangle = new ZRectangle();
        this.mExpandedInterfaceRectangle.setFillPaint(null);
        this.mExpandedInterfaceSeparator = new ZLine();

	this.addVisualComponent(mInterfaceLabel);
	this.addVisualComponent(mInterfaceEllipse);
	dataChanged();        
    }

    /** 
     * Add parameter to the list.
     * @param inElement
     */
    public void addSubsDeps(VModelElement inElement)
    {
	mSubsDeps.add(inElement);
    }
    /**
     * Clear all elements from substitute dependency list.
     */
    public void removeSubsDeps()
    {
	mSubsDeps.clear();
    }
    /** Sets parameter to hidden relationship list
     * @param inElement
     */
    public void setHiddenRelationships(VModelElement inElement){
	mHiddenRelationships.add(inElement);
    }
    /** Gets content of hidden relationship list.
     * @return mHiddenRelationships
     */
    public LinkedList getHiddenRelationships(){
	return mHiddenRelationships;
    }
    /**
     * Clear hidden relationship list.
     */
    public void clearHiddenRelationships(){
        mHiddenRelationships.clear();
    }
    /**
     * Gets content of substitute dependency list.
     * @return mSubsDeps
     */
    public LinkedList getSubsDeps()
    {
	return mSubsDeps;
    }

    public int getInterfaceType()
    {
	return mInterfaceType;
    }

    public Object[] getOperations()
    {
	return mOperations;
    }

    public int getViewMode()
    {
	return mViewMode;
    }

    public void setInterfaceType(int inType)
    {
	if (inType == mInterfaceType)
	    return;
	
	mInterfaceType = inType;
	dataChanged();
    }

    public void setOperations(Object[] inOps)
    {
	mOperations = inOps;
	dataChanged();
    }

    public void setViewMode(int inMode)
    {
	if (inMode == mViewMode)
	    return;
        
	mViewMode = inMode;

        // remove current & add new visual components

        clearVisualComponents();

        if (mViewMode == VIEWMODE_CIRCLE) {
            addVisualComponent(mInterfaceLabel);
            addVisualComponent(mInterfaceEllipse);
        } else if (mViewMode == VIEWMODE_EXPANDED) {
            addVisualComponent(mExpandedInterfaceStereoLabel);
            addVisualComponent(mExpandedInterfaceLabel);
            addVisualComponent(mExpandedInterfaceRectangle);
            addVisualComponent(mExpandedInterfaceOpsText);
            addVisualComponent(mExpandedInterfaceSeparator);
        }

	dataChanged();
    }

    protected void dataChanged()
    {
        if (mModelElementVisible == false)
            this.editor().getInvisibleGroup();
        else
            this.editor().removeInvisibleGroup();
        
    
        if (mViewMode == VIEWMODE_CIRCLE) {
            // view as circle
            mInterfaceLabel.setText(mModelElementName);

            if (mInterfaceType == TYPE_OFFERED)               
                mInterfaceEllipse.setFillPaint(Color.black);
            else
                mInterfaceEllipse.setFillPaint(Color.white);
            
            // label width is used to determine the point where to
            // draw the circle
            ZBounds myLabelBounds = mInterfaceLabel.getBoundsReference();
        
            // find position for the circle so it's centered to the label
            double myEllipseX = (myLabelBounds.getWidth() - CIRCLE_SIZE) / 2;
            mInterfaceEllipse.setFrame(myEllipseX,0,CIRCLE_SIZE, CIRCLE_SIZE);
            
            // find bounds (height) of the ellipse and position label below it
            ZBounds myEllipseBounds = mInterfaceEllipse.getBoundsReference();
            mInterfaceLabel.setTranslation(0, myEllipseBounds.getHeight() + 5);
        } else {
            // view as expanded
            mExpandedInterfaceLabel.setText(mModelElementName);
            mExpandedInterfaceStereoLabel.setText("<<interface>>");
            StringBuffer myOps = new StringBuffer();
            boolean myFirstAdded = false;
            for (int i = 0; i < mOperations.length; i++)
                if (mOperations[i] != null) {
                    if (myFirstAdded)
                        myOps.append("\n");
                    myOps.append(mOperations[i].toString());
                    myFirstAdded = true;
                }
            mExpandedInterfaceOpsText.setText(myOps.toString());

            ZBounds myLabelBounds = 
                mExpandedInterfaceLabel.getBoundsReference();
            ZBounds myStereoLabelBounds = 
                mExpandedInterfaceStereoLabel.getBoundsReference();
            ZBounds myOpsBounds =
                mExpandedInterfaceOpsText.getBoundsReference();

            // find out width of the widest text component

            double myClassWidth;

            myClassWidth = Math.max(myLabelBounds.width,
                                    myStereoLabelBounds.width);
            myClassWidth = Math.max(myClassWidth,
                                    myOpsBounds.width);

            // add spacing

            myClassWidth += 2 * 4.0;

            // calculate positions for visual components

            double myStereoX = 
                myClassWidth / 2.0 - myStereoLabelBounds.width / 2.0;
            double myStereoY = 4.0;

            double myLabelX = myClassWidth / 2.0 - myLabelBounds.width / 2.0;
            double myLabelY = myStereoY + myStereoLabelBounds.height + 4;
                
            double mySeparatorX1 = 0;
            double mySeparatorY1 = myLabelY + myLabelBounds.height + 4;
            double mySeparatorX2 = myClassWidth;
            double mySeparatorY2 = mySeparatorY1;

            double myOpsX = 4.0;
            double myOpsY = mySeparatorY1 + 4.0;

            double myRectX = 0.0;
            double myRectY = 0.0;
            double myRectW = myClassWidth;
            double myRectH = myOpsY + myOpsBounds.height + 4;

            mExpandedInterfaceStereoLabel.setTranslation(myStereoX, myStereoY);
            mExpandedInterfaceLabel.setTranslation(myLabelX, myLabelY);
            mExpandedInterfaceSeparator.setLine(mySeparatorX1, mySeparatorY1,
                                                mySeparatorX2, mySeparatorY2);
            mExpandedInterfaceOpsText.setTranslation(myOpsX, myOpsY);
            mExpandedInterfaceRectangle.setRect(myRectX, myRectY,
                                                myRectW, myRectH);
        }
    }
	
	public Point2D getRelationshipConnectPoint()
	{
		if(getViewMode() == VIEWMODE_CIRCLE) {
                    ZBounds myEllipseBounds = 
                        mInterfaceEllipse.getBounds();
                    
                    localToGlobal(myEllipseBounds);
                    
                    return myEllipseBounds.getCenter2D();
		} else {
                    ZBounds myRectBounds = 
                        mExpandedInterfaceRectangle.getBounds();

                    localToGlobal(myRectBounds);
                    return myRectBounds.getCenter2D();
		}
	}
	
	public Point2D intersectWithLine(Line2D inLine)
	{
		if(getViewMode() == VIEWMODE_CIRCLE)
		{
			Rectangle2D myLabelBounds = mInterfaceLabel.getBounds();
			Rectangle2D myEllipseBounds = mInterfaceEllipse.getBounds();
			
			localToGlobal(myLabelBounds);
			localToGlobal(myEllipseBounds);
			
			Point2D myLabelIntersectionPoint = Intersections.lineIntersectsRectangle(inLine, myLabelBounds);
			Point2D myEllipseIntersectionPoint = Intersections.lineIntersectsEllipse(inLine, myEllipseBounds);
			
			if(myLabelIntersectionPoint == null && myEllipseIntersectionPoint == null)
				return null;
			if(myLabelIntersectionPoint == null && myEllipseIntersectionPoint != null)
				return myEllipseIntersectionPoint;
			if(myLabelIntersectionPoint != null && myEllipseIntersectionPoint == null)
				return myLabelIntersectionPoint;
			
			// Use squared distances for greater performance
			double myLabelPointDistance = myLabelIntersectionPoint.distanceSq(inLine.getP2());
			double myEllipsePointDistance = myEllipseIntersectionPoint.distanceSq(inLine.getP2());
			
			if(myLabelPointDistance < myEllipsePointDistance)
				return myLabelIntersectionPoint;
			else
				return myEllipseIntersectionPoint;
		}
		else
		{
                    ZBounds myBounds = getGlobalBounds();
                    return Intersections.lineIntersectsRectangle(inLine, myBounds);
		}
	}
	
}








