package venice;

import edu.umd.cs.jazz.ZVisualLeaf;
import edu.umd.cs.jazz.ZGroup;
import edu.umd.cs.jazz.ZNode;
import java.util.LinkedList;
import java.util.ListIterator;

/**
 * ZVisualLeafModelElement is a common superclass for all different
 * visual components except ZPackage. This class includes model element
 * handling functions so we don't need to duplicate that code everywhere.
 *
 * @author Hannu Laurila
 */

public abstract class ZVisualLeafModelElement extends ZVisualLeaf
    implements VModelElement
{
    // VModelElement instance variables

    /** model element id */
    protected String mModelElementId;

    /** model element name */
    protected String mModelElementName;

    /** model element type */
    protected String mModelElementType;

    /** is this model element visible */
    protected boolean mModelElementVisible;
    
    /** The list of original relationships */
    private LinkedList mOriginalRelationships = null;

    /** List of this elements offered interface substituted realizations.*/
    private LinkedList mOfferedRealizationSubstitutions = null;
    /** List of this elements hidden offered interface realizations.*/
    private LinkedList mHiddenOfferedRealizations = null;
    /** List of this elements required interface realizations.*/
    private LinkedList mRequiredRealizations = null;

    private LinkedList mRequiredInterfaceSubstitutions = null;

    private LinkedList mCompositeRelationships;
    
    /**
     * dataChanged() is called when model element data is changed.
     *
     * subclasses should override this method with a method which
     * modifies the component appearance according to data changes.
     */
    protected abstract void dataChanged();
    
    public ZVisualLeafModelElement(String inId)
    {
	this.mModelElementId = inId;
	this.mModelElementName = "";
	this.mModelElementType = "";
	this.mModelElementVisible = true;
	this.mOfferedRealizationSubstitutions = new LinkedList();
	this.mHiddenOfferedRealizations = new LinkedList();
	this.mRequiredInterfaceSubstitutions = new LinkedList();
	this.mRequiredRealizations = new LinkedList();
	this.mCompositeRelationships = new LinkedList();
    }

    // VModelElement operations begin

    public String getId()
    {
	return this.mModelElementId;
    }

    public String getName()
    {
	return this.mModelElementName;
    }

    public String getType()
    {
	return this.mModelElementType;
    }

    public boolean getVisible()
    {
	return this.mModelElementVisible;
    }

    public void setName(String inName)
    {
	if (inName == null)
	    inName = "";

	this.mModelElementName = inName;
	this.dataChanged();
    }
	
	public void setType(String inType)
	{
		if (inType == null)
			inType = "";
		
		this.mModelElementType = inType;
		this.dataChanged();
	}

	public void setVisible(boolean inState)
	{
		if (inState == this.mModelElementVisible)
	    	// no change
			return;
		
		this.mModelElementVisible = inState;
		this.dataChanged();
    }

    public LinkedList getRelationships(VPackage inCrossThisPackage)
    {
	LinkedList myTempList = new LinkedList();
	ListIterator myIterator;
	
	if(mOriginalRelationships != null) {
	    myIterator = mOriginalRelationships.listIterator(0);
	    
	    while(myIterator.hasNext()) {
		ZRelationship myRel = (ZRelationship) myIterator.next();
		if(myRel.leavesVPackage(inCrossThisPackage) &&
		   !myRel.getIsRepresentedBySomeoneElse())
		{
		    myTempList.add(myRel);
		}
	    }
	}
	
	if(mCompositeRelationships != null) {
	    myIterator = mCompositeRelationships.listIterator(0);
	    
	    while(myIterator.hasNext())
	    {
		VRelationship myRel = (VRelationship) myIterator.next();
		
		if(myRel.leavesVPackage(inCrossThisPackage) &&
		   !myRel.getIsRepresentedBySomeoneElse())
		    {
			myTempList.add(myRel);
		    }
	    }
	}
	
	if(!mOfferedRealizationSubstitutions.isEmpty()){
	    myIterator = mOfferedRealizationSubstitutions.listIterator(0);
	    while(myIterator.hasNext())
	    {
		ZRelationship myRel = (ZRelationship) myIterator.next();
		if(myRel.leavesVPackage(inCrossThisPackage) &&
				   !myRel.getIsRepresentedBySomeoneElse())
		{
		    myTempList.add(myRel);
		}
	    }
	}
	if(!mRequiredInterfaceSubstitutions.isEmpty()){
	    myIterator = mRequiredInterfaceSubstitutions.listIterator(0);
	    while(myIterator.hasNext())
	    {
		ZRelationship myRel = (ZRelationship) myIterator.next();
		if(myRel.leavesVPackage(inCrossThisPackage) &&
		   !myRel.getIsRepresentedBySomeoneElse())
		{
		    myTempList.add(myRel);
		}
	    }
	}
	
	if(myTempList.isEmpty())
	    return null;
	else
	    return myTempList;
    }
    
    public void addRelationship(VRelationship inRelationship)
    {
	if(mOriginalRelationships == null)
	    mOriginalRelationships = new LinkedList();
	
	if(inRelationship.getSource() == this || 
	   inRelationship.getDestination() == this)
	    mOriginalRelationships.add(inRelationship);
    }
    
    /** Add relationship to this list.*/
    public void addToHiddenOfferedRealizations(VRelationship inRea)
    {
	mHiddenOfferedRealizations.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToOfferedRealizationSubstitutions(VRelationship inRea)
    {
	mOfferedRealizationSubstitutions.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToRequiredInterfaceSubstitutions(VRelationship inRea)
    {
	mRequiredInterfaceSubstitutions.add(inRea);
    }
    /** Add relationship to this list.*/
    public void addToRequiredRealizations(VRelationship inRea){
	mRequiredRealizations.add(inRea);
    }
    /** Clear content of this list.*/
    public void clearHiddenOfferedRealizations()
    {
	mHiddenOfferedRealizations.clear();
    }
    /** Clear content of this list.*/
    public void clearOfferedRealizationSubstitutions()
    {
	mOfferedRealizationSubstitutions.clear();
    }
    /** Clear inRea from this list.*/
    public void clearRequiredInterfaceSubstitutions(VRelationship inRea)
    {
	mRequiredInterfaceSubstitutions.remove(inRea);
    }
    /** Clear content of this list.*/
    public void clearRequiredRealizations(){
        mRequiredRealizations.clear();
    }

    /** Get content of this list.*/
    public LinkedList getRequiredRealizations(){
	return mRequiredRealizations;
    }
    /** Get content of this list.*/
    public LinkedList getHiddenOfferedRealizations()
    {
	return mHiddenOfferedRealizations;
    }
    /** Get content of this list.*/
    public LinkedList getOfferedRealizationSubstitutions()
    {
	return mOfferedRealizationSubstitutions;
    }
    /** Get content of this list.*/
    public LinkedList getRequiredInterfaceSubstitutions()
    {
	return mRequiredInterfaceSubstitutions;
    }


    public VPackage getVParent()
    {
        ZNode myPackage = this;
	
        // go up the tree until we find a ZPackage.
	
        do {
            myPackage = myPackage.getParent();
        } while ( (myPackage != null) && !(myPackage instanceof ZPackage) );
	
        if (myPackage != null && myPackage instanceof ZPackage)
            return (VPackage) myPackage;
        else
            return null;
    }
    
    public LinkedList getCompositeRelationships()
    {
	return mCompositeRelationships;	
    }
    
    public VRelationship addToCompositeRelationshipList(VRelationship inRel)
    {
	boolean isSourceInThisPackage;
	
	if(((ZNode) inRel.getSource()).isDescendentOf(this))
	    isSourceInThisPackage = true;
	else
	    isSourceInThisPackage = false;
	
	inRel.setIsRepresentedBySomeoneElse(true);
	ListIterator myIterator = mCompositeRelationships.listIterator(0);
	
	while(myIterator.hasNext())
	{
	    ZRelationship myZRelationship = (ZRelationship) myIterator.next();
	    
	    // Source is in this package, so only type, relationship type
	    // and destination have to match. Source doesn't have to match,
	    // because source is this in the composite relationship.
	    if(isSourceInThisPackage)
	    {
		if(myZRelationship.getType() == inRel.getType() &&
		   myZRelationship.getDestination() == inRel.getDestination()
		   && myZRelationship.getRelationshipType() ==
		   ((ZRelationship) inRel).getRelationshipType())
		{
		    myZRelationship.addVRelationshipAsSubrelationship(inRel);
		    return null;
		}
	    }
	    else
	    {
		if(myZRelationship.getType() == inRel.getType() &&
		   myZRelationship.getSource() == inRel.getSource() &&
		   myZRelationship.getRelationshipType() ==
		   ((ZRelationship) inRel).getRelationshipType())
		{
		    myZRelationship.addVRelationshipAsSubrelationship(inRel);
		    return null;
		}
		
	    }
	}
	
	// No matching composite relationship found, must create a new one.
	
	ZRelationship myZRelationship = null;
	
	// FIXME!! Use ZVisualization to create an unique ID
	if(inRel instanceof ZRealization)
	    myZRelationship = new ZRealization(null, true);
	else if(inRel instanceof ZGeneralization)
	    myZRelationship = new ZGeneralization(null, true);
	else if(inRel instanceof ZDependency)
	    myZRelationship = new ZDependency(null, true);
	
	myZRelationship.setType(inRel.getType());
	myZRelationship.setName("Composite relationship");
	
	if(isSourceInThisPackage)
	{
	    myZRelationship.setSource(this);
	    myZRelationship.setDestination(inRel.getDestination());
	}
	else
	{
	    myZRelationship.setSource(inRel.getSource());
	    myZRelationship.setDestination(this);
	}
	
	inRel.setIsRepresentedBySomeoneElse(true);
	myZRelationship.addVRelationshipAsSubrelationship(inRel);
	
	mCompositeRelationships.add(myZRelationship);
	
	ZVisualization myVis = ZVisualization.getZVisualizationInstance();
	
	myVis.addVRelationshipWithoutAddingOriginalReference(myZRelationship);
	
	return myZRelationship;
    }

    public void debugDump()
    {
	VRelationship myRel = null;
	
	ListIterator it = mCompositeRelationships.listIterator();
	
	while(it.hasNext()) {
	    myRel = (VRelationship) it.next();
	    System.out.println("comp rel src " + myRel.getSource().getId() + " dst " + myRel.getDestination().getId() + " " + myRel.getIsRepresentedBySomeoneElse());
	}
    }
    
    // VModelElement operations end
}


