package venice;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;

import edu.umd.cs.jazz.*;
import edu.umd.cs.jazz.util.*;
import edu.umd.cs.jazz.event.*;
import edu.umd.cs.jazz.component.*;

/**
 * ZVisualization is a concrete implementation of Visualization interface.
 * <p>
 * ZVisualization uses Jazz zoomable user interface library as its backend.
 *
 * @author Hannu Laurila
 * @author Antti Pietarinen
 * @author Petteri Kamppuri
 */
public class ZVisualization implements Visualization
{
    /**
     * if this is set true, all relationships are put to the root package.
     * It makes things faster but the visual look is not as good when
     * they relationships are put to deeper level packages.
     * This speedup is not needed much anymore, because of optimizations
     * we made to Jazz library.
     */
    private static final boolean RELATIONSHIP_SPEEDUP_ENABLED = false;

    /** how much space is left to edges when zooming to an element */
    private static final double ZOOMSPACING = 50;

    /** animation length for zoomTo */
    private static final int ZOOMTIME = 500;

    /** cursor which is used when visualization area has no active mode */
    private static Cursor stNoneCursor =
        Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

    /** cursor which is used in panning mode */
    private static Cursor stPanCursor =
        Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR);

    /** cursor which is used in select mode */
    private static Cursor stSelectCursor =
        Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

    /** cursor which is used in zooming mode */
    private static Cursor stZoomCursor =
        Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);

    /** jazz canvas which is a Swing component which can be placed to UI */
    private ZCanvas mCanvas;

    /** reference to the currently active event handler */
    private ZEventHandler mCurrentEventHandler;

    /** event handler which is active in panning mode */
    private ZPanEventHandler mPanEventHandler;

    /** event handler which is active in zooming mode */
    private VZoomEventHandler mZoomEventHandler;

    /** event handler which is active in selection mode */
    private VCompositeSelectionHandler mSelectionHandler;

    /** selection layer is used to draw the selection marquee */
    private ZLayerGroup mSelectionLayer;

    /**
     * Hashtable<String,VModelElement> for 
     * quickly accessing vmodelelements 
     */
    private Hashtable mIndexById;

    /** the root VPackage in which everything from the model is contained */
    private ZPackage mRootPackage;

    /** a listener (e.g. the GUI) which is notified on selection changes */
    private VSelectionListener mSelectionListener;

    /** current mode of the visualization (one of Visualization.MODE...) */
    private int mCurrentMode;

    /** reference to instantiated visualization */
    protected static ZVisualization stVisualizationReference = null;

    private boolean mInInteractiveMode;

    /**
     * constructor which sets up all internal data structures.
     */
    public ZVisualization()
    {
        mInInteractiveMode = true;

        // create canvas (the ui component)
        mCanvas = new ZCanvas();

        // Create a selection layer for drawing the marquee
        mSelectionLayer = new ZLayerGroup();
        mCanvas.getRoot().addChild(mSelectionLayer);
        mCanvas.getCamera().addLayer(mSelectionLayer);

        // initialize event handlers and set current operation mode
        initEventHandlers();
        setToolMode(MODE_NONE);

        mIndexById = new Hashtable();

        // create the root package (do not put it to index)
        mRootPackage = new ZPackage("__ID__RootPackage__",
                                    mInInteractiveMode);
        mRootPackage.setName("Root Package");
        mRootPackage.setRootPackage(true);

        // put the root package to the visualization
        mCanvas.getLayer().addChild(mRootPackage);

        if(stVisualizationReference != null)
            System.err.println("ZVisualization instantiated twice!\nThis will generate problems.");

        stVisualizationReference = this;

        // Listen for selection changes in the visualization.
        ZSelectionManager.addGroupListener(new ZGroupListener() {
                public void nodeAdded(ZGroupEvent e) {
                    fireSelectionChanged();
                }
                public void nodeRemoved(ZGroupEvent e) {
                    fireSelectionChanged();
                }
            });
        
        /*
        mCanvas.getDrawingSurface().setRenderQuality(ZDrawingSurface.RENDER_QUALITY_LOW);
        mCanvas.getDrawingSurface().setInteractingRenderQuality(ZDrawingSurface.RENDER_QUALITY_LOW);
        */
    }
    
    /**
     * Returns the current instance of ZVisualization.
     */
    public static ZVisualization getZVisualizationInstance()
    {
        return stVisualizationReference;
    }

    /**
     * Adds a relationship to the visualization. 
     * Relationships is either a realization, a generalization
     * or a dependency. The relationship is added to a list
     * of original relationships in both source and destination node.
     * That information is used in various <code>Commands</code> to
     * perform for example relationship grouping on parent level.
     * This means that the given <code>VRelationship</code> must have
     * its source and destination set.
     * 
     * @param inRelship relationship to add to visualization.
     */
    public void addVRelationship(VRelationship inRelship)
    {
        addVRelationshipWithoutAddingOriginalReference(inRelship);


        if(!(inRelship.getSource() instanceof VRelationship))
            inRelship.getSource().addRelationship(inRelship);
        if(!(inRelship.getDestination() instanceof VRelationship))
            inRelship.getDestination().addRelationship(inRelship);

    }

    /**
     * Adds the given <code>VRelationship</code> to the visualization,
     * but does not add it as an original relationship to the source and
     * destination nodes.
     * 
     * @param inRelship relationship to add to visualization
     */
    public void addVRelationshipWithoutAddingOriginalReference(VRelationship inRelship)
    {

        if(inRelship.getSource() == null ||
                inRelship.getDestination() == null) {
            System.err.println("addVRelationship: inRelship doesn't " +
                               "have source or destination. " +
                               "Relation not added.");
            return;
        }
	addToIndex(inRelship);
        if (!(inRelship instanceof ZRelationship)) {
            // FIXME: ERROR
            return;
        }

        if (RELATIONSHIP_SPEEDUP_ENABLED) {
            // put relationship to the root package
            getRootPackage().addVChild(inRelship);
            return;
        }


        ZNode myCommonParent = (ZNode) inRelship.getSource();
        ZNode myDst = (ZNode) inRelship.getDestination();

        do {
            if (myDst.isDescendentOf(myCommonParent) &&
                    myCommonParent instanceof ZPackage)
                break;

            myCommonParent = myCommonParent.getParent();
        } while (myCommonParent != null);

        // if no common root was found use the root package
        if (myCommonParent == null) {
            mRootPackage.addVChild(inRelship);
	    // FIXME  dataChanged here is a partial hack
	    ((ZRelationship) inRelship).dataChanged();
            return;
        }

        // common parent package was found
        ((ZPackage)myCommonParent).addVChild(inRelship);
	
	// FIXME  dataChanged here is a partial hack
	((ZRelationship) inRelship).dataChanged();
    }

    public VModelElement findElementById(String inId)
    {
        return (VModelElement) this.mIndexById.get(inId);
    }

    public JMenu getDebugMenu()
    {
        JMenu myMenu = new JMenu("JAZZ Visualization");

        JMenuItem myTreeDumpItem =
            new JMenuItem("Dump Scenegraph to stderr");
        JMenuItem myTest1DataItem =
            new JMenuItem("Put test model 1 to the root package");
        JMenuItem myTest2DataItem =
            new JMenuItem("Put test model 2 to the root package");
        JMenuItem myDefaultLayoutItem =
            new JMenuItem("Make default layout");
        JMenuItem myClearItem =
            new JMenuItem("Clear visualization");
        JMenuItem myHideItem =
            new JMenuItem("Hide for two seconds");
        JMenuItem myContentItem =
            new JMenuItem("Toggle showContent");
        JMenuItem myViewModeItem =
            new JMenuItem("Toggle viewMode");
        JMenuItem myTreeViewItem =
            new JMenuItem("Open scenegraph tree view");
        JMenuItem mySurpriseItem =
            new JMenuItem("Today's surprise debug menu item");

        myMenu.add(myTreeDumpItem);
        myMenu.add(myTest1DataItem);
        myMenu.add(myTest2DataItem);
        myMenu.add(myDefaultLayoutItem);
        myMenu.add(myClearItem);
        myMenu.add(myHideItem);
        myMenu.add(myContentItem);
        myMenu.add(myViewModeItem);
        myMenu.add(myTreeViewItem);
        myMenu.add(mySurpriseItem);

        myTreeDumpItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    ZDebug.dump(mCanvas.getRoot());
                }
            });
        
        myTest1DataItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    setupTestModel1();
                }
            });
        
        myTest2DataItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    setupTestModel2();
                }
            });
        
        myDefaultLayoutItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    makeDefaultLayout();
                }
            });
        
        myClearItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    clear();
                }
            });
        
        myHideItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    VModelElement[] mySel = getSelection();
                    
                    if (mySel == null)
                        return;
                    
                    final VModelElement myElem = mySel[0];
                    
                    myElem.setVisible(false);
                    
                    final Runnable myRestorer = new Runnable() {
                            public void run() {
                                myElem.setVisible(true);
                            }
                        };
                    
                    (new Thread() {
                            public void run() {
                                try {
                                    Thread.sleep(2000);
                                }
                                catch (InterruptedException ex) {
                                }
                                SwingUtilities.invokeLater(myRestorer);
                            }
                        }).start();
                }
            });
        
        myContentItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    VModelElement[] mySel = getSelection();
                    
                    if (mySel == null)
                        return;
                    
                    final VModelElement myElem = mySel[0];
                    
                    if (myElem instanceof VPackage) {
                        VPackage myPackage = (VPackage) myElem;
                        myPackage.setShowContent(!myPackage.getShowContent());
                    }
                }
            });
        
        myViewModeItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    VModelElement[] mySel = getSelection();
                    
                    if (mySel == null)
                        return;
                    
                    final VModelElement myElem = mySel[0];
                    
                    if (myElem instanceof VInterface) {
                        VInterface myInterface = (VInterface) myElem;
                        if (myInterface.getViewMode() == VInterface.VIEWMODE_CIRCLE)
                            myInterface.setViewMode(VInterface.VIEWMODE_EXPANDED);
                        else
                            myInterface.setViewMode(VInterface.VIEWMODE_CIRCLE);
                        
                    }
                }
            });
        
        myTreeViewItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    ZSceneGraphTreeView treeView = new ZSceneGraphTreeView(mCanvas);
                    treeView.pack();
                    treeView.show();
                }
            });

        mySurpriseItem.addActionListener( new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    mCanvas.getDrawingSurface().printSurface();
                }
            });
        
        return myMenu;
    }
    
    public VPackage getRootPackage()
    {
        return mRootPackage;
    }

    public VModelElement[] getSelection()
    {
        // load selection to mElements

        ArrayList mySelection =
            ZSelectionManager.getSelectedNodes(mCanvas.getRoot());
        VModelElement[] mElements = new VModelElement[mySelection.size()];

        // System.out.println("\nCurrent selection:");
        for(int i = 0; i< mySelection.size(); i++){
            mElements[i]= (VModelElement) mySelection.get(i);
            // System.out.println("\t" + mElements[i].getName());
        }

        if(mElements.length > 0)
            return mElements;
        else
            return null;
    }

    public JComponent getSwingComponent()
    {
        return mCanvas; // ZCanvas returned as JComponent
    }

    public int getToolMode()
    {
        return mCurrentMode;
    }

    public void makeDefaultLayout()
    {
        mRootPackage.makeDefaultLayout(true, 1);
        zoomTo(mRootPackage);
    }

    public VComponent newVComponent(String inId)
    {
        VComponent myNew = new ZComponent(inId);
        addToIndex(myNew);
        return myNew;
    }

    public VInterface newVInterface(String inId)
    {
        VInterface myNew = new ZInterface(inId);
        addToIndex(myNew);
        return myNew;
    }

    public VPackage newVPackage(String inId)
    {
        VPackage myNew = new ZPackage(inId, mInInteractiveMode);
        addToIndex(myNew);
        return myNew;
    }

    public VDependency newVDependency(String inId)
    {
        VDependency myNew = new ZDependency(inId, mInInteractiveMode);
        addToIndex(myNew);
        return myNew;
    }

    public VGeneralization newVGeneralization(String inId)
    {
        VGeneralization myNew = new ZGeneralization(inId, mInInteractiveMode);
        addToIndex(myNew);
        return myNew;
    }

    public VRealization newVRealization(String inId)
    {
        VRealization myNew = new ZRealization(inId, mInInteractiveMode);
        addToIndex(myNew);
        return myNew;
    }

    public void removeElement(String inId)
    {
        removeFromIndex(inId);
    }

    public void removeVRelationship(VRelationship inRelship)
    {
	// FIXME!!
	//((ZPackage) ((ZRelationship) inRelship).getParent()).removeVChild(inRelship);
	//getRootPackage().removeVChild(inRelship);
	//removeElement(inRelship.getId());
	
	// removeVRelationshipFromPackage(getRootPackage(), inRelship);
	
	VPackage myRelationshipParent = inRelship.getVParent();
	
	if(myRelationshipParent != null)
	    myRelationshipParent.removeVChild(inRelship);
    }

    /**
     * Removes the relationship from this package. If the relationship doesn't
     * belong to this package, this method will try to remove the relationship
     * from this package's children.
     * Returns whether the relationship was removed.
     *
     * 
     * @param inPackage the from which to remove the relationship
     * @param inRel the relationships to remove
     * @return whether relationships was removed
     */
    private boolean removeVRelationshipFromPackage(VPackage inPackage, VRelationship inRel)
    {
	VModelElement[] myChildren = inPackage.getVChildren();
	
	for(int i = 0; i < myChildren.length; ++i) {
	    if(myChildren[i] == inRel) {
		inPackage.removeVChild(inRel);
		return true;
	    }
	    else if(myChildren[i] instanceof VPackage) {
		if(removeVRelationshipFromPackage((VPackage) myChildren[i], inRel))
		    return true;
	    }
	}
	
	return false;
    }
    
    public void saveLayout(OutputStream inoutStream) throws IOException
    {
        // Unselect everything in the view. There is a bug in Jazz
        // which causes NullPointerException if there are selected
        // nodes in the scenegraph when serializing.

        VModelElement[] mySavedSelection = getSelection();
        setSelection(null);

        SerializedZVisualization mySerVis = new SerializedZVisualization();

        mySerVis.mRootPackage = mRootPackage;
        mySerVis.mIndexById = mIndexById;

        // remove root package from visualization temporalily
        // this prevents saving unnecessary nodes in the srlzed file.
        mCanvas.getLayer().removeChild(mRootPackage);

        ObjectOutputStream myOos = new ObjectOutputStream(inoutStream);
        myOos.writeObject(mySerVis);
        myOos.close();

        // re-add root package
        mCanvas.getLayer().addChild(mRootPackage);

        // re-select stuff which was selected before operation
        setSelection(mySavedSelection);
    }

    public void loadLayout(InputStream inoutStream) throws IOException
    {
        ObjectInputStream myOis = new ObjectInputStream(inoutStream);
        SerializedZVisualization mySerVis = null;

        try {
            mySerVis = (SerializedZVisualization) myOis.readObject();
        }
        catch (Exception myEx) {
            myEx.printStackTrace();
            throw new IOException("Loading layout failed.");
        }
        myOis.close();

        // remove old root package from visualization
        mCanvas.getLayer().removeChild(mRootPackage);

        mRootPackage = mySerVis.mRootPackage;
        mIndexById = mySerVis.mIndexById;

        // add new root package
        mCanvas.getLayer().addChild(mRootPackage);

        // re-enable listeners for nodes recursively
        // (listener states are not saved in serialized form)

        if (mInInteractiveMode)
            mRootPackage.setListenersEnabled(true,true);

        // fit to screen
        zoomTo(mRootPackage);
    }


    public void setSelection(VModelElement[] inSelection)
    {
        ZSelectionManager.unselectAll(mRootPackage);

        if (inSelection == null)
            return;

        for (int i = 0; i < inSelection.length; i++)
            ZSelectionManager.select((ZNode) inSelection[i]);
    }

    public void setSelectionListener(VSelectionListener inListener)
    {
        this.mSelectionListener = inListener;
    }

    public void setToolMode(int inMode)
    {
        // if no event handler was enabled, don't try to disable.
        if (mCurrentEventHandler != null)
            mCurrentEventHandler.setActive(false);

        mCanvas.setNavEventHandlersActive(false);

        switch(inMode)
        {
        case MODE_PAN:
            mCurrentMode = MODE_PAN;
            mCurrentEventHandler = mPanEventHandler;
            mCanvas.setCursor(stPanCursor);
            break;
        case MODE_SELECT:
            mCurrentMode = MODE_SELECT;
            mCurrentEventHandler = mSelectionHandler;
            mCanvas.setCursor(stSelectCursor);
            break;
        case MODE_ZOOM:
            mCurrentMode = MODE_ZOOM;
            mCurrentEventHandler = mZoomEventHandler;
            mCanvas.setCursor(stZoomCursor);
            break;
        default:
            // fall back to MODE_NONE
            mCurrentMode = MODE_NONE;
            mCurrentEventHandler = null;
            mCanvas.setCursor(stNoneCursor);
            break;
        }

        // don't try to enable the MODE_NONE event handler
        if (mCurrentEventHandler != null)
            mCurrentEventHandler.setActive(true);
    }

    public void setZoomVolume(int inDx)
    {
        mZoomEventHandler.setDx(inDx*2);
    }

    public void zoomTo(VModelElement inElement)
    {
        if (inElement == null)
            return;

        ZNode myNode = (ZNode) inElement;

        zoomTo(myNode.getGlobalBounds());
    }

    public void zoomToSelection()
    {
        zoomTo(getSelectionBounds());
    }

    /**
     * zooms to specified bounds. if bounds width or height is zero,
     * returns without zooming to avoid odd effects.
     */
    private void zoomTo(ZBounds myTargetBounds)
    {
        if (myTargetBounds == null)
            return;

        if ( (myTargetBounds.getWidth() == 0) ||
                (myTargetBounds.getHeight() == 0) )
            return;

        ZBounds myCameraBounds = mCanvas.getCamera().getBounds();
        myCameraBounds.x = myCameraBounds.x + ZOOMSPACING;
        myCameraBounds.y = myCameraBounds.y + ZOOMSPACING;
        myCameraBounds.width = myCameraBounds.width - (ZOOMSPACING * 2);
        myCameraBounds.height = myCameraBounds.height - (ZOOMSPACING * 2);

        mCanvas.getCamera().center(myTargetBounds,
                                   myCameraBounds,
                                   ZOOMTIME,
                                   mCanvas.getCamera().getDrawingSurface());
    }

    /**
     * calculates bounds for current selection.
     *
     * Returns null if there is no selection.
     */
    private ZBounds getSelectionBounds()
    {
        VModelElement[] mySelection = getSelection();
        if (mySelection == null)
            return null;

        ZBounds mySelectionBounds = new ZBounds();
        mySelectionBounds.reset();
        for (int i = 0; i < mySelection.length; i++)
            mySelectionBounds.add( ((ZNode)mySelection[i]).getGlobalBounds());

        return mySelectionBounds;
    }


    public void clear()
    {
        mRootPackage.removeAllVChildren(true); // recurse = true
        this.mIndexById = new Hashtable();
        fireSelectionChanged();
    }

    private void fireSelectionChanged()
    {
        // if there is a listener, notify it
        // System.out.println("firing selectionChanged!");
        if (mSelectionListener != null)
            mSelectionListener.selectionChanged();
    }

    /**
     * Initialize the event handlers
     */
    private void initEventHandlers()
    {
        ZNode myCameraNode = mCanvas.getCameraNode();
        ZCamera myCamera = mCanvas.getCamera();

        mPanEventHandler =
            new ZPanEventHandler(myCameraNode,myCamera);
        mZoomEventHandler =
            new VZoomEventHandler(myCameraNode, myCamera);
        mSelectionHandler =
            new VCompositeSelectionHandler(myCameraNode,
                                           mCanvas,
                                           mSelectionLayer);

        // disable deletion handler (objects can not be deleted using del key)
        mSelectionHandler.setEnabled(VCompositeSelectionHandler.DELETE,
                                     false);
    }

    /**
     * sets up an test model for testing visualization subsystem and
     * different graphical components.
     */
    private void setupTestModel1()
    {
        VPackage myRoot = getRootPackage();

        VComponent vc1 = newVComponent("Component1");
        vc1.setName("ProducerComponent");
        myRoot.addVChild(vc1);

        VComponent vc2 = newVComponent("Component2");
        vc2.setName("ConsumerComponent");
        myRoot.addVChild(vc2);

        VPackage vp3 = newVPackage("Package1");
        vp3.setName("FunnyPackage");
        myRoot.addVChild(vp3);

        VComponent vc4 = newVComponent("Component4");
        vc4.setName("JokeComponent");
        vp3.addVChild(vc4);

        VComponent vc5 = newVComponent("Component5");
        vc5.setName("SatireComponent");
        vp3.addVChild(vc5);

        VPackage vp6 = newVPackage("Package6");
        vp6.setName("CartoonPackage");
        vp3.addVChild(vp6);

        VComponent vc7 = newVComponent("Component7");
        vc7.setName("UserFriendlyComponent");
        vp6.addVChild(vc7);

        VComponent vc8 = newVComponent("Component8");
        vc8.setName("DilbertComponent");
        vp6.addVChild(vc8);

        VInterface vi9 = newVInterface("Interface9");
        vi9.setInterfaceType(VInterface.TYPE_OFFERED);
        vi9.setOperations(new String[] {
                              "getName()",
                              "getPriority()",
                              "join()",
                              "resume()",
                              "sleep()"
                          });

        vi9.setName("OfferedInterface");
        vp6.addVChild(vi9);

        VInterface vi10 = newVInterface("Interface10");
        vi10.setName("RequiredInterface");
        vi10.setInterfaceType(VInterface.TYPE_REQUIRED);
        vp3.addVChild(vi10);

        VPackage vp11 = newVPackage("Package11");
        vp11.setName("BoyPackage");
        myRoot.addVChild(vp11);

        VPackage vp12 = newVPackage("Package12");
        vp12.setName("GirlPackage");
        myRoot.addVChild(vp12);

        VDependency vd13 = newVDependency("Dependency13");
        vd13.setType("needs (dep)");
        vd13.setSource(vp11);
        vd13.setDestination(vp12);
        addVRelationship(vd13);

        VRealization vr14 = newVRealization("Realization14");
        vr14.setType("realization");
        vr14.setSource(vc8);
        vr14.setDestination(vi9);
        addVRelationship(vr14);

        VGeneralization vg15 = newVGeneralization("Generalization15");
        vg15.setType("generalization");
        vg15.setSource(vc7);
        vg15.setDestination(vc8);
        addVRelationship(vg15);

        this.makeDefaultLayout();
    }

    /**
     * sets up another test model for testing the subsystem
     */
    private void setupTestModel2()
    {
        VPackage myRoot = getRootPackage();

        VComponent vc1 = newVComponent("1");
        vc1.setName("ComponentOne");

        VComponent vc2 = newVComponent("2");
        vc2.setName("ComponentTwo");

        myRoot.addVChild(vc1);
        myRoot.addVChild(vc2);

        makeDefaultLayout();
    }

    /**
     * adds a VModelElement to the indexing hashtable with checks
     */
    private void addToIndex(VModelElement inElement)
    {
        if (inElement == null)
            return;

        String id = inElement.getId();

        if (id == null)
            return;

        this.mIndexById.put(id, inElement);
    }

    private void removeFromIndex(String outElement)
    {
        if(outElement == null)
            return;
        this.mIndexById.remove(outElement);
    }

    public void enterNonInteractiveMode()
    {
        if (mInInteractiveMode == true) {
            mInInteractiveMode = false;
            mRootPackage.setListenersEnabled(false, true);
        }
    }

    public void leaveNonInteractiveMode()
    {
        if (mInInteractiveMode == false) {
            this.mInInteractiveMode = true;
            mRootPackage.setListenersEnabled(true, true);
        }
    }
    
    public void printCurrentView()
    {
        mCanvas.getDrawingSurface().printSurface();
    }
}
